VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Level"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private Const MAXWIDTH = 170
Private Const MAXHEIGHT = 170

'Level information
Private mWidth As Integer
Private mHeight As Integer
Private LevelData() As Integer
Private InfoData() As Integer
Private mLevelNumber As Integer
Private mEpisode As Integer

'Default tiles for level border and background
Private BorderTile(0 To 2) As Integer
Private BlankTile(0 To 2) As Integer

Public Sub Create(Width As Integer, Height As Integer, Episode As Integer, LevelNumber As Integer)
    Dim X As Integer, Y As Integer
    
    mWidth = Width
    mHeight = Height
    ReDim LevelData(mWidth * mHeight) As Integer
    ReDim InfoData(mWidth * mHeight) As Integer
    For Y = 0 To mHeight - 1
        For X = 0 To mWidth - 1
            If X < 2 Or X >= mWidth - 2 Or Y < 2 Or Y >= mHeight - 2 Then
                LevelData(Y * mWidth + X) = BorderTile(Episode - 1)
            Else
                LevelData(Y * mWidth + X) = BlankTile(Episode - 1)
            End If
        Next
    Next
    mEpisode = Episode
    mLevelNumber = LevelNumber
End Sub

Public Property Let Width(ByVal newWidth As Integer)
    Resize newWidth, mHeight
End Property
Public Property Let Height(ByVal newHeight As Integer)
    Resize mWidth, newHeight
End Property

Public Property Get Width() As Integer
    Width = mWidth
End Property
Public Property Get Height() As Integer
    Height = mHeight
End Property

Public Sub Load(Filename As String)
    Dim fin As Integer, DataLen As Long, mFileTitle As String
    Dim j As Integer, b As Integer, n As Integer, i As Integer, X As Integer, Y As Integer
    
    i = InStr(Filename, "\")
    mFileTitle = Mid(Filename, i + 1)
    mEpisode = Val(Right(mFileTitle, 1))
    mLevelNumber = Val(Mid(mFileTitle, 6, 2))
    
    fin = FreeFile()
    Open Filename For Binary Access Read As #fin
    
    'Read the width and height
    Seek #fin, 4 + 1
    Get #fin, , mWidth
    Get #fin, , mHeight
        
    ReDim LevelData(mWidth * mHeight) As Integer
    ReDim InfoData(mWidth * mHeight) As Integer
    
    'Work out the padded length of layer data
    DataLen = ((mWidth * mHeight + 7) \ 8) * 8 + 8
    
    'Start reading the level data
    Seek #fin, 20 + 1
    j = 0: X = 0: Y = 0
    Do
        Get #fin, , b
        If EOF(1) Then
            Exit Do
        End If
        If b = &HFEFE Then
            Get #fin, , n
            Get #fin, , b
            For i = 0 To n - 1
                j = j + 1
                If j > 8 Then
                    If Y < mHeight Then
                        LevelData(Y * mWidth + X) = b
                        X = X + 1
                        If X = mWidth Then
                            X = 0
                            Y = Y + 1
                            If Y = mHeight * 2 Then Exit Do
                        End If
                    Else
                        If j > DataLen Then
                            InfoData((Y - mHeight) * mWidth + X) = b
                            X = X + 1
                            If X = mWidth Then
                                X = 0
                                Y = Y + 1
                                If Y = mHeight * 2 Then Exit Do
                            End If
                        End If
                    End If
                End If
            Next
        Else
            j = j + 1
            If j > 8 Then
                If Y < mHeight Then
                    LevelData(Y * mWidth + X) = b%
                    X = X + 1
                    If X = mWidth Then
                        X = 0
                        Y = Y + 1
                        If Y = mHeight * 2 Then Exit Do
                    End If
                Else
                    If j > DataLen Then
                        InfoData((Y - mHeight) * mWidth + X) = b
                        X = X + 1
                        If X = mWidth Then
                            X = 0
                            Y = Y + 1
                            If Y = mHeight * 2 Then Exit Do
                        End If
                    End If
                End If
            End If
        End If
    Loop
    Close #fin
End Sub

Public Sub Display(ByRef Dest As PictureBox, ByVal sx As Integer, ByVal sy As Integer, ByRef Tiles As TileSet, Sprites As SpriteSet, Optional ByVal odx As Integer = 0, Optional ByVal ody As Integer = 0, Optional ByVal w As Integer = -1, Optional ByVal h As Integer = -1)
    Dim X As Integer, Y As Integer, s As Integer
    Dim dx As Integer, dy As Integer
    
    'Calculate the size of the picturebox in tiles
    If w = -1 Then w = Tiles.PixelWToTiles(Dest.ScaleWidth)
    If h = -1 Then h = Tiles.PixelHToTiles(Dest.ScaleHeight)
    
    'Draw the tiles
    dx = odx: dy = ody
    For Y = sy To sy + h - 1
        For X = sx To sx + w - 1
            If Y < mHeight And X < mWidth Then
                Tiles.Draw Tile(X, Y), Dest, dx, dy
            Else
                Tiles.Draw -1, Dest, dx, dy
            End If
            dx = dx + Tiles.RenderWidth
        Next
        dy = dy + Tiles.RenderHeight
        dx = odx
    Next
    
    'Draw the sprites -- extend a bit further than the visible area so we see those
    'sprites which originate off the edges
    dx = -3 * Tiles.RenderWidth + odx: dy = -3 * Tiles.RenderHeight + ody
    For Y = sy - 3 To sy + h - 1
        For X = sx - 3 To sx + w - 1
            If Y >= 0 And Y < mHeight And X >= 0 And X < mWidth Then
                s = InfoData(Y * mWidth + X)
                Select Case s
                    Case 255:   'Keen or MapKeen
                        'Draw mapkeen if it's a map level
                        If mLevelNumber = 80 Then
                            Sprites.Draw -1, Dest, dx, dy
                        Else
                            Sprites.Draw 0, Dest, dx, dy
                        End If
                    Case 1 To Sprites.Count - 1:    'Sprites
                        Sprites.Draw s, Dest, dx, dy
                    Case Else:  'Switch, Level, or Teleporter
                End Select
            End If
            dx = dx + Tiles.RenderWidth
        Next
        dy = dy + Tiles.RenderHeight
        dx = -3 * Tiles.RenderWidth + odx
    Next
End Sub

Public Sub Save(Filename As String)
    Dim fin As Integer
    Dim DataLen As Long, MemSize As Long
    Dim s As String, i As Integer
    
    fin = FreeFile
    Open Filename For Binary Access Write As #fin

    'Length of plane data with padding
    DataLen = ((mWidth * mHeight + 7) \ 8) * 8
    MemSize = DataLen * 4 + 32  '32 bytes in level header

    Put #fin, , CLng(MemSize)
    Put #fin, , CInt(mWidth)
    Put #fin, , CInt(mHeight)
    s = Chr(2) + String(5, 0)
    Put #fin, , s
    s = Chr(0) + Chr(0) + Chr(0) + Chr(0)
    Put #fin, , s
    Put #fin, , CInt(DataLen * 2)
    s = Chr(&HFE) + Chr(&HFE) + Chr(8) + Chr(0) + Chr(0) + Chr(0)
    Put #fin, , s

    'Store compressed level data
    CompressData LevelData(), mWidth, mHeight, fin

    'Put padding in end of plane data
    For i = 1 To DataLen - (mWidth * mHeight)
        s = Chr(0) + Chr(0)
        Put #fin, , s
    Next

    'Store compressed info plane data
    CompressData InfoData(), mWidth, mHeight, fin

    'Put padding in end of plane data
    For i = 1 To DataLen - (mWidth * mHeight)
        s = Chr(0) + Chr(0)
        Put #fin, , s
    Next
    
    Close #fin
End Sub

Private Sub CompressData(IData() As Integer, w As Integer, h As Integer, filen As Integer)
    Dim k As String * 2
    k = Chr$(&HFE) + Chr$(&HFE) 'RLE Magic Word
    
    Dim last As Integer, lastn As Integer, b As Integer
    Dim X As Integer, Y As Integer, i As Integer

    last = -1
    lastn = 0
    For Y = 0 To h - 1
        For X = 0 To w - 1
            b = IData(Y * w + X)
            If lastn > 0 Then
                If b <> last Then
                    If lastn >= 4 Then
                        Put #filen, , k
                        Put #filen, , lastn
                        Put #filen, , last
                        last = b
                        lastn = 1
                    Else
                        For i = 1 To lastn
                            Put #filen, , last
                        Next
                        last = b
                        lastn = 1
                    End If
                    lastn = 1
                    last = b
                Else
                    lastn = lastn + 1
                End If
            Else
                last = b
                lastn = 1
            End If
        Next
    Next
    If lastn > 0 Then
        If lastn >= 4 Then
            Put #filen, , k
            Put #filen, , lastn
            Put #filen, , last
        Else
            For i = 1 To lastn
                Put #filen, , last
            Next
        End If
        lastn = 1
        last = b
    End If
End Sub

Public Property Let Tile(ByVal X As Integer, ByVal Y As Integer, ByVal newTile As Integer)
    If X >= 0 And X < mWidth And Y >= 0 And Y < mHeight Then
        LevelData(Y * mWidth + X) = newTile
    End If
End Property

Public Property Get Tile(ByVal X As Integer, ByVal Y As Integer) As Integer
    If X >= 0 And X < mWidth And Y >= 0 And Y < mHeight Then
        Tile = LevelData(Y * mWidth + X)
    Else
        Tile = 0
    End If
End Property

Public Property Let Info(ByVal X As Integer, ByVal Y As Integer, ByVal newInfo As Integer)
    If X >= 0 And X < mWidth And Y >= 0 And Y < mHeight Then
        InfoData(Y * mWidth + X) = newInfo
    End If
End Property

Public Property Get Info(ByVal X As Integer, ByVal Y As Integer) As Integer
    If X >= 0 And X < mWidth And Y >= 0 And Y < mHeight Then
        Info = InfoData(Y * mWidth + X)
    Else
        Info = 0
    End If
End Property

Public Sub Resize(newWidth As Integer, newHeight As Integer)
    Dim X As Integer, Y As Integer
    'Create temporary arrays to hold the stuff that stays
    ReDim tmpLevelData(newWidth * newHeight) As Integer
    ReDim tmpInfoData(newWidth * newHeight) As Integer
    
    'Copy the old data into it
    For Y = 0 To newHeight - 1
        For X = 0 To newWidth - 1
            If X < mWidth And Y < mHeight Then
                tmpLevelData(Y * newWidth + X) = LevelData(Y * mWidth + X)
                tmpInfoData(Y * newWidth + X) = InfoData(Y * mWidth + X)
            Else
                If X < 2 Or X >= newWidth - 2 Or Y < 2 Or Y >= newHeight - 2 Then
                    tmpLevelData(Y * newWidth + X) = BorderTile(Episode - 1)
                Else
                    tmpLevelData(Y * newWidth + X) = BlankTile(Episode - 1)
                End If
                tmpInfoData(Y * newWidth + X) = 0
            End If
        Next
    Next
    
    'Resize the original arrays
    mWidth = newWidth
    mHeight = newHeight
    ReDim LevelData(mWidth * mHeight) As Integer
    ReDim InfoData(mWidth * mHeight) As Integer
    
    For Y = 0 To mHeight - 1
        For X = 0 To mWidth - 1
            LevelData(Y * mWidth + X) = tmpLevelData(Y * mWidth + X)
            InfoData(Y * mWidth + X) = tmpInfoData(Y * mWidth + X)
        Next
    Next
End Sub

Private Sub Class_Initialize()
    BorderTile(0) = 144
    BorderTile(1) = 144
    BorderTile(2) = 245
    BlankTile(0) = 143
    BlankTile(1) = 143
    BlankTile(2) = 143
End Sub

Public Property Let Episode(newEpisode As Integer)
    mEpisode = newEpisode
End Property

Public Property Get Episode() As Integer
    Episode = mEpisode
End Property


Public Property Let Number(newNumber As Integer)
    mLevelNumber = newNumber
End Property

Public Property Get Number() As Integer
    Number = mLevelNumber
End Property


